/*
 * Copyright (c) 1992, 1996
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#ifndef lint
char copyright[] =
    "@(#) Copyright (c) 1992, 1996\n\
The Regents of the University of California.  All rights reserved.\n";
static  char rcsid[] =
    "@(#)$Header: gpstime.c,v 1.8 96/07/04 19:25:32 leres Exp $ (LBL)";
#endif

#include <sys/types.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/ppsclock.h>

#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <xntp/ntp_calendar.h>
#include <xntp/ntp_unixtime.h>

#include "gpstime.h"
#include "mx4200.h"
#include "io.h"
#include "util.h"

u_long	caltontp __P((struct calendar *));

static int quick;

static char *gps = "/dev/gps0";

static const char pmvxg[] = "PMVXG";

extern char *optarg;

static	void gpsconfig __P((int));
static	void gpsloop __P((int));
static	int gpsrequest __P((int, int, char *, u_int, u_int));
static	int gpsgetev __P((int, struct ppsclockev *));
static	char *fmttv __P((struct timeval *));
static	int gpsparse __P((char *, int *, struct calendar *));

int
main(argc, argv)
	int argc;
	char **argv;
{
	register char *cp;
	register int op;
	register int fd;
	int hack = 0;

	if ((cp = strrchr(argv[0], '/')) != NULL)
		prog = cp + 1;
	else
		prog = argv[0];
	while ((op = getopt(argc, argv, "dvqhg:")) != EOF)
		switch(op) {

		case 'd':
			++debug;
			break;

		case 'v':
			++verbose;
			break;

		case 'q':
			++quick;
			break;

		case 'h':
			++hack;
			break;

		case 'g':
			gps = optarg;
			break;

		default:
			(void)fprintf(stderr,
			    "usage: %s [-dvqh] [-g gpsdev]\n", prog);
			exit(1);
		}

	/* Open the gps serial port */
	if ((fd = open(gps, O_RDWR, 0)) < 0) {
		fprintf(stderr, "%s: fopen ", prog);
		perror(gps);
		exit(1);
	}

	/* Setup proper baud rate and terminal characteristics */
	if (!gpsinit(fd))
		exit(1);

	/* Configure */
	if (!quick)
		gpsconfig(fd);

	if (hack)
		for (;;) {
			struct ppsclockev ev;

			(void)gpsgetev(fd, &ev);
			printf("%d.%06d\t%d\n",
			    ev.tv.tv_sec, ev.tv.tv_usec, ev.serial);
		}
	
	/* Time keeping loop */
	gpsloop(fd);
	exit(0);
}

void
gpsconfig(fd)
	register int fd;
{
	char buf[1024];

	/* Zero the output list */
	gpssend(fd, "%s,%03d,,%d,,,,,,", pmvxg, PMVXG_S_PORTCONF,
	    1);			/* clear current output list */

	/* Turn off time recovery */
	gpssend(fd, "%s,%03d,,,,,,%d,", pmvxg, PMVXG_S_TRECOVCONF,
	    0);			/* do not output time recovery messages */

	/* Request software configuration */
	while (!gpsrequest(fd, PMVXG_D_SOFTCONF, buf, sizeof(buf), 3))
		continue;
	fprintf(stderr, "software config: \"%s\"\n", buf);

	/* Request status */
	while (!gpsrequest(fd, PMVXG_D_STATUS, buf, sizeof(buf), 3))
		continue;
	fprintf(stderr, "status: \"%s\"\n", buf);

	/* Switch to 2d mode */
	gpssend(fd, "%s,%03d,%d,,%.1f,%.1f,,%d,%d,%c,%d",
	    pmvxg, PMVXG_S_INITMODEB,
	    2,			/* 2d mode */
	    0.1,		/* hor accel fact as per Steve */
	    0.1,		/* ver accel fact as per Steve */
	    10,			/* hdop limit as per Steve */
	    5,			/* elevation limit as per Steve */
	    'U',		/* time output mode */
	    0);			/* local time offset from gmt */

	/* Request mode data */
	while (!gpsrequest(fd, PMVXG_D_MODEDATA, buf, sizeof(buf), 3))
		continue;
	fprintf(stderr, "mode data: \"%s\"\n", buf);

	/* Set position */
#ifdef notdef
	gpssend(fd, "%s,%03d,,,,,%1.3f,%c,%1.3f,%c,%5.1f,",
	    pmvxg, PMVXG_S_INITMODEA,
	    37.8772, 'N', 122.2487, 'W',/* LBL via xmap */
	    275.0);			/* Jef: About 250 - 300 meters */
#endif

	/* Request position */
	while (!gpsrequest(fd, PMVXG_D_POSITION, buf, sizeof(buf), 3))
		continue;
	fprintf(stderr, "position: \"%s\"\n", buf);

	/* Configure time recovery */
	gpssend(fd, "%s,%03d,%c,%c,%c,%d,%d,%d,", pmvxg, PMVXG_S_TRECOVCONF,
#ifdef notdef
	    'K',		/* known position */
	    'D',		/* dynamic position */
#else
	    'S',		/* static position */
#endif
	    'U',		/* steer clock to gps time */
	    'A',		/* always output time pulse */
	    100,		/* max time error in ns */
	    0,			/* user bias in ns */
	    1);			/* output to control port */

	/* Request time recovery usage */
	while (!gpsrequest(fd, PMVXG_D_TRECOVUSEAGE, buf, sizeof(buf), 3))
		continue;
	fprintf(stderr, "time recovery usage: \"%s\"\n", buf);
}

static char *months[13] = {
	"???",
	"Jan",
	"Feb",
	"Mar",
	"Apr",
	"May",
	"Jun",
	"Jul",
	"Aug",
	"Sep",
	"Oct",
	"Nov",
	"Dec"
};

char *
xtime(timep)
const time_t * const    timep;
{
	return asctime(gmtime(timep));
}


static void
gpsloop(fd)
	register int fd;
{
	register int n;
	char buf[1024];
	int valid;
	struct ppsclockev ev;
	struct timeval tv;
	register struct calendar *jt;
	struct calendar sjt;
	long long x1, x2, x;
	u_int lastserial = 0;

	jt = &sjt;
	fprintf(stderr, "gpsloop: entered\n");
	for (;;) {
		n = gpsgets(fd, buf, sizeof(buf), 10);
		if (n == 0) {
			fprintf(stderr, "gpsloop: timeout\n");

			/* Are you (still) there? */
			n = 4;
			for (n = 4; !gpsrequest(fd, PMVXG_D_STATUS,
			    buf, sizeof(buf), 2); --n)
				continue;
			fprintf(stderr, "status: %s\n", buf);
			if (n > 0)
				continue;

			/* Reinitialize */
			(void) gpseatme(fd);
			gpsconfig(fd);
			continue;
		}
		(void) gpsgetev(fd, &ev);
		/* Ascii message is actually for *next* pulse */
		ev.tv.tv_sec++;
		fprintf(stderr, "gpsloop: Unix: %15s", fmttv(&ev.tv));
		if (gpsparse(buf, &valid, jt)) {
			/* Print gps time as recovered */
			fprintf(stderr, " GPS: %2d:%02d:%02d %2d-%s-%04d",
			    jt->hour, jt->minute, jt->second,
			    jt->monthday, months[jt->month], jt->year);
			if (!valid)
				fprintf(stderr, " (pulse not valid)");

			/* Convert to Unix time */
			tv.tv_sec = caltontp(jt);
			tv.tv_sec -= JAN_1970;
			tv.tv_usec = 0;
			if (verbose)
				fprintf(stderr, " (%.24s)", xtime(&tv.tv_sec));

			/* Print delta */
			x1 = ((long long)ev.tv.tv_sec) * 1000000 +
			    (long long)ev.tv.tv_usec;
			x2 = ((long long)tv.tv_sec) * 1000000 +
			    (long long)tv.tv_usec;
			x = x1 - x2;
			if (x < 0)
				fprintf(stderr, " (d: -%d.%06d)",
				    (int)(-x / 1000000), (int)(-x % 1000000));
			else
				fprintf(stderr, " (d: %d.%06d)",
				    (int)x / 1000000, (int)x % 1000000);
			if (lastserial + 1 != ev.serial && lastserial != 0)
				fprintf(stderr, " (%d)",
				ev.serial - lastserial);
			lastserial = ev.serial;
		}
		if (debug > 1)
			fprintf(stderr, " \"%s\"", buf);
		fputc('\n', stderr);
	}
}

static int
gpsrequest(fd, t, buf, size, timeout)
	register int fd, t;
	char *buf;
	register u_int size, timeout;
{
	register int i, n;
	register char *cp;

	/* Send the request */
	gpssend(fd, "%s,%03d", "CDGPQ", t);

	for (n = 3; n > 0; --n) {
		/* Wait for the response */
		if (gpsgets(fd, buf, size, timeout) == 0) {
			fprintf(stderr, "%s: gpsrequest: timeout\n", prog);
			return (0);
		}
		if ((cp = strchr(buf, ',')) == NULL) {
			fprintf(stderr, "%s: gpsrequest: missing 2nd arg\n",
			    prog);
			return (0);
		}
		i = atoi(++cp);
		if (i == t)
			return (1);
	}
	fprintf(stderr, "gpsrequest: wrong type %d != %d\n", i, t);
	return (0);
}

static int
gpsgetev(fd, evp)
	register int fd;
	register struct ppsclockev *evp;
{
	if (ioctl(fd, CIOGETEV, (char *)evp) < 0) {
		perror("pps ioctl");
		evp->tv.tv_sec = -1;
		evp->tv.tv_usec = 0;
		evp->serial = 0;
		return (0);
	}
	return (1);
}

static char *
fmttv(tvp)
	register struct timeval *tvp;
{
	register struct tm *tm;
	static char buf[32];
	time_t t;

	if (tvp->tv_sec == -1 && tvp->tv_usec == -1)
		return ("?");

	t = tvp->tv_sec;
	tm = gmtime(&t);
	if (tm == NULL)
		return ("!@#$%");
#ifdef notdef
	/* BSD is rational */
	(void) snprintf(buf, sizeof(buf), "%2d:%02d:%02d.%06d",
	    tm->tm_hour, tm->tm_min, tm->tm_sec, tvp->tv_usec);
#else
	/* SunOS sucks */
	(void) sprintf(buf, "%2d:%02d:%02d.%06d",
	    tm->tm_hour, tm->tm_min, tm->tm_sec, tvp->tv_usec);
#endif
	return (buf);
}

static int
gpsparse(buf, validp, jt)
	register char *buf;
	register int *validp;
	register struct calendar *jt;
{
	register char *cp;
	register int i;

	cp = buf;
	memset((char *)jt, 0, sizeof(*jt));

	if ((cp = strchr(cp, ',')) == NULL)
		return (0);
	i = atoi(++cp);
	if (i != PMVXG_D_TRECOVOUT)
		return (0);

	if ((cp = strchr(cp, ',')) == NULL)
		return (0);
	if (*++cp == 'T')
		*validp = 1;
	else if (*cp == 'F')
		*validp = 0;
	else
		return (0);

	/* Year */
	if ((cp = strchr(cp, ',')) == NULL)
		return (0);
	i = atoi(++cp);
	jt->year = i;

	/* Month of year */
	if ((cp = strchr(cp, ',')) == NULL)
		return (0);
	i = atoi(++cp);
	jt->month = i;

	/* Day of month */
	if ((cp = strchr(cp, ',')) == NULL)
		return (0);
	i = atoi(++cp);
	jt->monthday = i;

	/* Hour */
	if ((cp = strchr(cp, ',')) == NULL)
		return (0);
	i = atoi(++cp);
	jt->hour = i;

	/* Minute */
	if ((cp = strchr(cp, ':')) == NULL)
		return (0);
	i = atoi(++cp);
	jt->minute = i;

	/* Second */
	if ((cp = strchr(cp, ':')) == NULL)
		return (0);
	i = atoi(++cp);
	jt->second = i;

	return (1);
}


